"""Integration test for multiple tags."""
import random
import sys
from copy import deepcopy

import pytest
import pytest_check as check

from cliconfig import (
    Config,
    create_processing_keep_property,
    create_processing_value,
    make_config,
)
from cliconfig.process_routines import merge_flat_processing


def test_multiple_tags() -> None:
    """Integration test for multiple tags."""
    sys_argv = sys.argv.copy()
    sys.argv = ["test_inte_multiple_tags.py"]
    config = make_config("tests/configs/integration/test1/main.yaml")
    expected_config = {
        "path_1": "tests/configs/integration/test1/sub1.yaml",
        "path_2": "tests/configs/integration/test1/sub2.yaml",
        "config1": {
            "param": 2,
            "param2": 1,
        },
        "config2": {
            "param": 2,
        },
        "config3": {
            "select": "config3.param1",
            "param1": 0,
        },
    }
    check.equal(config.dict, expected_config)
    with pytest.raises(
        ValueError,
        match="Key previously tagged with '@type:None|int'.*"
    ):
        merge_flat_processing(
            config,
            Config({'config2.param': 5.6}, []),
            preprocess_first=False,
        )
    sys.argv = sys_argv


def test_multiple_tags2() -> None:
    """2nd integration test for multiple tags."""
    sys_argv = sys.argv.copy()
    sys.argv = [
        "test_inte_multiple_tags2.py",
        "--config",
        "[tests/configs/integration/test2/exp1.yaml, "
        "tests/configs/integration/test2/exp2.yaml]",
        "--train.n_epochs=20",
        "--train.optimizer.momentum=0",
        "--train.optimizer.type=Adam",
    ]

    def func_pos_enc_type(x: str) -> str:
        if x in ['absolute', 'relative', 'embed']:
            return x
        raise ValueError(f"Invalid value for pos_enc_type: {x}")

    def func_optim_type(x: str) -> str:
        if x in ['SGD', 'Adam']:
            return x
        raise ValueError(f"Invalid value for optim_type: {x}")
    proc_pos_enc_type = create_processing_value(
        func_pos_enc_type,
        tag_name='pos_enc',
        order=20,
        persistent=True,
    )
    proc_optim_type = create_processing_value(
        func_optim_type,
        tag_name='optim_type',
        order=20,
        persistent=True,
    )
    proc_protect = create_processing_keep_property(
        func=lambda x: x,
        tag_name='protect',
        premerge_order=-15,
        postmerge_order=15,
    )
    proc_run_id = create_processing_value(
        lambda run_id: run_id if run_id is not None else random.randint(0, 1000000),
        regex="run_id.*",
    )
    config = make_config(
        "tests/configs/integration/test2/default.yaml",
        process_list=[proc_pos_enc_type, proc_optim_type, proc_protect, proc_run_id],
    )
    expected_dict = {
        "project_name": "ImageClassif",
        "models": {
            "archi_name": "models.vit_b16",
            "vit_b16": {
                "pos_enc_type": "absolute",
                "attn_dropout": 0.2,
                "dropout": 0.1,
                "in_size": 512,
                "n_blocks": 12,
            },
        },
        "data": {
            "data_size": 512,
            "dataset_path": "../../mydata",
            "standardization": True,
            "augmentation": ["RandomHorizontalFlip", "RandomVerticalFlip"],
            "dataset_cfg_path": "tests/configs/integration/test2/data.yaml",
        },
        "train": {
            "n_epochs": 20,
            "optimizer": {
                "type": "Adam",
                "lr": 0.01,
                "momentum": 0,
            },
        },
    }
    check.is_instance(config.dict["run_id"], int)
    config_dict = deepcopy(config.dict)
    del config_dict["run_id"]
    check.equal(config_dict, expected_dict)
    sys.argv = sys_argv



"""Tests for dict routines with preprocessing."""
import re
import shutil

import pytest
import pytest_check as check
import yaml

from cliconfig.base import Config
from cliconfig.process_routines import (
    load_processing,
    merge_flat_paths_processing,
    merge_flat_processing,
    save_processing,
)
from cliconfig.processing.base import Processing
from tests.conftest import ProcessAdd1, ProcessKeep


def test_merge_flat_processing(
    process_add1: ProcessAdd1,
    process_keep: ProcessKeep,
) -> None:
    """Test merge_flat_processing."""
    class _ProcessingTest(Processing):

        def __init__(self) -> None:
            super().__init__()
            self.attr = 0

    proc1 = _ProcessingTest()
    proc2 = _ProcessingTest()
    proc2.attr = 1
    proc3 = _ProcessingTest()

    config1 = Config({"a": {"b": 1, "c": 2, "d@keep": 3}}, [proc1, proc2, process_add1])
    config2 = Config({"a": {"b": 2, "c@add1": 3, "d": 4}}, [proc3, process_keep])
    config = merge_flat_processing(
        config1,
        config2,
        allow_new_keys=False,
    )
    check.equal(
        config.dict,
        {"a.b": 2, "a.c": 4, "a.d": 3},
    )
    check.equal(config.process_list, [proc1, proc2, process_add1, process_keep])
    check.equal(process_keep.keep_vals, {})


def test_merge_flat_paths_processing(
    process_add1: ProcessAdd1,
    process_keep: ProcessKeep,
) -> None:
    """Test merge_flat_paths_processing."""
    config1 = Config(
        {"param1@add1": 0, "param2.param3@keep": 1},
        [process_add1, process_keep]
    )
    config2 = Config({"param2.param3": 3}, [])
    expected_dict = {"param1": 1, "param2.param3": 1}
    check.equal(
        merge_flat_paths_processing(
            "tests/configs/configtag1.yaml",
            "tests/configs/configtag2.yaml",
            allow_new_keys=False,
            additional_process=[process_add1, process_keep],
        ),
        Config(expected_dict, [process_add1, process_keep]),
    )
    check.equal(
        merge_flat_paths_processing(
            config1,
            config2,
            allow_new_keys=False,
        ),
        Config(expected_dict, [process_add1, process_keep]),
    )
    check.equal(process_keep.keep_vals, {})

    # Case with dict in input


def test_save_processing(
    process_add1: ProcessAdd1,
    process_keep: ProcessKeep,
) -> None:
    """Test save_processing."""
    config = Config(
        {"param1@add1": 0, "param2.param3@add1": 1},
        [process_add1, process_keep]
    )
    save_processing(config, "tests/tmp/config.yaml")
    with open("tests/tmp/config.yaml", "r", encoding="utf-8") as yaml_file:
        loaded_dict = yaml.safe_load(yaml_file)
    check.equal(loaded_dict, {"param1": 1, "param2": {"param3": 2}})
    check.equal(process_keep.keep_vals, {})
    shutil.rmtree("tests/tmp")


def test_load_processing(
    process_add1: ProcessAdd1,
    process_keep: ProcessKeep,
) -> None:
    """Test load_processing."""
    process_keep.keep_vals = {"param2.param3": 0}
    config = load_processing(
        "tests/configs/configtag2.yaml",
        [process_add1, process_keep],
    )
    check.equal(config.dict, {"param2.param3": 0})
    with pytest.raises(
        ValueError,
        match=re.escape(
            "config_or_path must be a Config instance or a path to a yaml file "
            "but you passed a dict. If you want to use it as a valid input, "
            "you should use Config(<input dict>, []) instead."
        )
    ):
        merge_flat_paths_processing(
            {"a": 2},  # type: ignore
            Config({"a": 1}, []),
        )
    with pytest.raises(
        ValueError,
        match=("config_or_path must be a Config instance or a path to a yaml file.")
    ):
        merge_flat_paths_processing(
            Config({"a": 1}, []),
            ("not a path",),  # type: ignore
        )



"""Built-in processing classes.

Classes to apply pre-merge, post-merge, pre-save and post-load modifications
to dict with processing routines (found in cliconfig.process_routines).
"""
from typing import Any, Dict, List, Set

from cliconfig.base import Config
from cliconfig.process_routines import (
    merge_flat_paths_processing,
    merge_flat_processing,
)
from cliconfig.processing._type_parser import _isinstance, _parse_type
from cliconfig.processing.base import Processing
from cliconfig.tag_routines import clean_all_tags, clean_tag, dict_clean_tags, is_tag_in


class ProcessMerge(Processing):
    """Merge dicts just in time with '@merge_after/_before/_add' tags.

    Tag your key with '@merge_after', '@merge_before' or @merge_add to load
    the dict corresponding to the value (path) and merge it just before or after
    the current dict. The merged dicts will be processed with pre-merge but
    not post-merge to ensure that merged configurations are all processed with
    pre-merge before applying a post-merge processing. It allows the merged
    configs to make references to each other (typically for copy).

    * '@merge_add' merges the dict corresponding to the path by allowing ONLY new keys
      It is a security check when you want to add a dict completely new
      It is a typical usage for a default config splitted in several files.
    * '@merge_after' merge the dict corresponding to the path on the current dict
    * '@merge_before' merge the current dict on the dict corresponding to the path

    The processing is a pre-merge processing only and occurs before
    most of the other processing.
    Pre-merge order: -20.0

    Examples
    --------
    .. code-block:: yaml

        --- # config1.yaml
        a:
          b: 1
          b_path@merge_after: dict2.yaml
        --- # config2.yaml
        a.b: 2
        c_path@merge_add: config3.yaml
        --- # config3.yaml
        c: 3

    Before merging, the config1 is interpreted as the dict:

    ::

        {'a': {'b': 2, 'b_path': 'config2.yaml'}, 'c_path': 'config3.yaml', 'c': 3}`

    If you replace '@merge_after' by '@merge_before', it will be:

    ::

        {'a': {'b': 1, 'b_path': 'config2.yaml'}, 'c_path': 'config3.yaml', 'c': 3}`

    Finally, if you replace '@merge_after' by '@merge_add', it will raises an
    error because the key 'a.b' already exists in the dict.
    """

    def __init__(self) -> None:
        super().__init__()
        self.premerge_order = -20.0

    def premerge(self, flat_config: Config) -> Config:
        """Pre-merge processing."""
        items = list(flat_config.dict.items())
        for flat_key, val in items:
            if is_tag_in(flat_key, "merge_after"):
                if not isinstance(val, str) or not val.endswith('.yaml'):
                    raise ValueError(
                        "Key with '@merge_after' tag must be associated "
                        "to a string corresponding to a *yaml* file."
                        f"The problem occurs at key: {flat_key}"
                    )
                # Remove the tag in the dict
                del flat_config.dict[flat_key]
                flat_config.dict[clean_tag(flat_key, 'merge_after')] = val
                # Merge + process the dicts
                # NOTE: we allow new keys with security because the merge
                # following this pre-merge will avoid the creation of
                # new keys if needed.
                flat_config = merge_flat_paths_processing(
                    flat_config,
                    val,
                    allow_new_keys=True,
                    preprocess_first=False,  # Already processed
                    postprocess=False,
                )

            elif is_tag_in(flat_key, "merge_before"):
                if not isinstance(val, str) or not val.endswith('.yaml'):
                    raise ValueError(
                        "Key with '@merge_before' tag must be associated "
                        "to a string corresponding to a *yaml* file."
                        f"The problem occurs at key: {flat_key}"
                    )
                # Remove the tag in the dict
                del flat_config.dict[flat_key]
                flat_config.dict[clean_tag(flat_key, 'merge_before')] = val
                # Merge + process the dicts
                flat_config = merge_flat_paths_processing(
                    val,
                    flat_config,
                    allow_new_keys=True,
                    preprocess_second=False,  # Already processed
                    postprocess=False,
                )

            elif is_tag_in(flat_key, "merge_add"):
                if not isinstance(val, str) or not val.endswith('.yaml'):
                    raise ValueError(
                        "Key with '@merge_add' tag must be associated "
                        "to a string corresponding to a *yaml* file."
                        f"The problem occurs at key: {flat_key}"
                    )
                # Remove the tag in the dict
                del flat_config.dict[flat_key]
                flat_config.dict[clean_tag(flat_key, 'merge_add')] = val
                # Pre-merge process the dict with the process list of
                # the current config
                flat_config_to_merge = merge_flat_paths_processing(
                    Config({}, []),
                    val,
                    additional_process=flat_config.process_list,
                    allow_new_keys=True,
                    preprocess_first=False,  # Already processed
                    postprocess=False,
                )
                clean_dict, _ = dict_clean_tags(flat_config.dict)
                clean_dict_to_merge, _ = dict_clean_tags(flat_config_to_merge.dict)
                for key in clean_dict_to_merge:
                    if clean_all_tags(key) in clean_dict:
                        raise ValueError(
                            f"@merge_add doest not allow to add already "
                            f"existing keys but key '{key}' is found in both "
                            "dicts. Use @merge_after or @merge_before if you "
                            "want to merge this key, or check your key names."
                        )
                # Merge the dicts (order is not important by construction)
                # NOTE: we delete the process list of the current config
                # to speed up the process by avoiding redundant processing
                flat_config = merge_flat_processing(
                    Config(flat_config.dict, []),
                    flat_config_to_merge,
                    allow_new_keys=True,
                    preprocess_first=False,  # Already processed
                    preprocess_second=False,  # Already processed
                    postprocess=False,
                )
        return flat_config


class ProcessCopy(Processing):
    """Copy a value with '@copy' tag. The copy is protected from updates.

    Tag your key with '@copy' and with value the name of the flat key to copy.
    Then, the value will be a copy of the corresponding value forever.
    The pre-merge processing will remove the tag. The post-merge processing
    will set the value (if the copied key exists) and occurs after most processing.
    The pre-save processing restore the tag and the key to copy to keep the
    information on future loads.

    The copy key is protected against any modification and will raise an error
    if you try to modify it.

    Pre-merge order: 0.0
    Post-merge order: 10.0
    Pre-save order: 10.0

    Examples
    --------
    .. code-block:: yaml

        # config.yaml
        a:
          b: 1
          c@copy: a.b

    Before merging, the config is interpreted as the dict

    .. code-block:: python

        {'a': {'b': 1, 'c': 1}}

    Note
    ----
        If the key to copy does not exist in the config on post-merge, the
        processing will NOT raise an error to let the user the possibility
        to add the key later via merge. However, the key still be protected.
    """

    def __init__(self) -> None:
        super().__init__()
        self.premerge_order = 0.0
        self.postmerge_order = 10.0
        self.presave_order = 10.0
        self.keys_to_copy: Dict[str, str] = {}
        self.current_value: Dict[str, Any] = {}

    def premerge(self, flat_config: Config) -> Config:
        """Pre-merge processing."""
        items = list(flat_config.dict.items())
        for flat_key, val in items:
            if is_tag_in(flat_key, "copy"):
                if not isinstance(val, str):
                    raise ValueError(
                        "Key with '@copy' tag must be associated "
                        "to a string corresponding to a flat key. "
                        f"The problem occurs at key: {flat_key} with value: {val}"
                    )
                clean_key = clean_all_tags(flat_key)
                if (clean_key in self.keys_to_copy
                        and self.keys_to_copy[clean_key] != val):
                    raise ValueError(
                        "Key with '@copy' has change its value to copy. Found key: "
                        f"{flat_key} with value: {val}, previous value to copy: "
                        f"{self.keys_to_copy[clean_key]}"
                    )
                # Store the key to copy and value
                self.keys_to_copy[clean_key] = val
                self.current_value[clean_key] = val
                # Remove the tag and update the dict
                flat_config.dict[clean_tag(flat_key, "copy")] = val
                del flat_config.dict[flat_key]
        return flat_config

    def postmerge(self, flat_config: Config) -> Config:
        """Post-merge processing."""
        for key, val in self.keys_to_copy.items():
            # NOTE: Do not raise an error if the key to copy does not exist
            # yet because it can be added later in a future merge
            if key in flat_config.dict and val in flat_config.dict:
                if flat_config.dict[key] != self.current_value[key]:
                    # The key has been modified
                    raise ValueError(
                        "Found attempt to modify a key with '@copy' tag. The key "
                        "is then protected against updates (except the copied "
                        f"value or the original key to copy). Found key: {key} of "
                        f"value {flat_config.dict[key]} that copy {val} of value "
                        f"{flat_config.dict[val]}")
                # Copy the value
                flat_config.dict[key] = flat_config.dict[val]
                # Update the current value
                self.current_value[key] = flat_config.dict[val]
        return flat_config

    def presave(self, flat_config: Config) -> Config:
        """Pre-save processing."""
        # Restore the tag with the key to copy to keep the information
        # on further loading
        for key, val in self.keys_to_copy.items():
            if key in flat_config.dict:
                new_key = key + "@copy"
                del flat_config.dict[key]
                flat_config.dict[new_key] = val
        return flat_config


class ProcessTyping(Processing):
    """Force a type with '@type:mytype' tag. The type is preserved forever.

    Allow basic types (none, any, bool, int, float, str, list, dict), nested lists,
    nested dicts, unions (with Union or the '|' symbol) and Optional.
    It store the type in pre-merge and check alls forced types on post-merge.
    It restore the tag in pre-save to keep the information on future loads.
    It always occurs last in post-merge.
    Pre-merge order: 0.0
    Post-merge order: 20.0
    Pre-save order: 0.0

    Note
    ----
        The type is not checked on pre-merge to allow the parameter to be
        updated (by a copy or a merge for instance). The goal of this
        processing is to ensure the type at the end of the post-merge.

    Examples
    --------
    ::

        dict1 = {param@type:None|List[int|float]: None}
        dict2 = {param: [0, 1, 2.0]}  # no error
        dict3 = {param: [0, 1, 2.0, 'a']}  # error

    Merge configs with dictionnaries dict1 and dict2 raise no error and `param`
    is forced to be None or a list of int or float forever. Merge dict1
    and dict3 raise an error on post-merge because of the 'a' value.

    Note that removing "None|" in the type description still doesn't raise an error
    in the first case because the type checking is evaluated after the merge with
    dict2.
    """

    def __init__(self) -> None:
        super().__init__()
        self.premerge_order = 0.0
        self.postmerge_order = 20.0
        self.presave_order = 0.0
        self.forced_types: Dict[str, tuple] = {}
        self.type_desc: Dict[str, str] = {}  # For error messages

    def premerge(self, flat_config: Config) -> Config:
        """Pre-merge processing."""
        items = list(flat_config.dict.items())
        for flat_key, val in items:
            end_key = flat_key.split('.')[-1]
            if "@type:" in end_key:
                # Get the type description
                trail = end_key.split("@type:")[-1]
                type_desc = trail.split('@')[0]  # (in case of multiple tags)
                expected_type = tuple(_parse_type(type_desc))
                clean_key = clean_all_tags(flat_key)
                if (clean_key in self.forced_types
                        and set(self.forced_types[clean_key]) != set(expected_type)):
                    raise ValueError(
                        f"Find the tag '@type:{type_desc}' on a key that has already "
                        "been associated to an other type: "
                        f"{self.type_desc[clean_key]}. "
                        f"Find problem at key: {flat_key}"
                    )
                # Remove the tag
                del flat_config.dict[flat_key]
                flat_config.dict[clean_tag(flat_key, f'type:{type_desc}')] = val
                # Store the forced type
                self.forced_types[clean_key] = expected_type
                self.type_desc[clean_key] = type_desc
        return flat_config

    def postmerge(self, flat_config: Config) -> Config:
        """Post-merge processing."""
        for key, expected_type in self.forced_types.items():
            if (key in flat_config.dict
                    and not _isinstance(flat_config.dict[key], expected_type)):
                type_desc = self.type_desc[key]
                raise ValueError(
                    f"Key previously tagged with '@type:{type_desc}' must be "
                    f"associated to a value of type {type_desc}. Find the "
                    f"value: {flat_config.dict[key]} of type "
                    f"{type(flat_config.dict[key])} at key: {key}"
                )
        return flat_config

    def presave(self, flat_config: Config) -> Config:
        """Pre-save processing."""
        # Restore the tag with the type to keep the information
        # on further loading
        for key, val in self.type_desc.items():
            if key in flat_config.dict:
                new_key = key + f"@type:{val}"
                flat_config.dict[new_key] = flat_config.dict[key]
                del flat_config.dict[key]
        return flat_config


class ProcessSelect(Processing):
    """Select a sub-config with and delete the rest of its parent config.

    First look in pre-merge for a parameter tagged with '@select' containing a
    flat key corresponding to a sub-configurations to keep. The parent configuration
    is then deleted on post-merge, except the selected sub-configuration
    and eventually the tagged parameter (if it is in the same sub-configuration).
    It is also possible to select multiple keys of a same sub-configuration
    (meaning that the part before the last dot must be equal) by passing a
    list of flat keys.
    Pre-merge order: 0.0
    Post-merge order: 0.0

    Examples
    --------
    .. code-block:: yaml

        models:
            model_names@select: [models.model1, models.model3]
            model1:
                param1: 1
                param2: 2
            model2:
                param1: 3
                param2: 4
            model3:
                submodel:
                    param: 5
            model4:
                param: 6

    Result in deleting models.model2 (param1 and param2) and models.model4.param,
    and keeping the rest.

    Warning
    -------

        For security reasons, it prevents from deleting the configuration at the root
        (which is the case when the selected key doesn't contain a dot),
        and raises an error in this case.

    """

    def __init__(self) -> None:
        super().__init__()
        self.keys_that_select: Set[str] = set()
        self.subconfigs_to_delete: Set[str] = set()
        self.keys_to_keep: Set[str] = set()

    def premerge(self, flat_config: Config) -> Config:
        """Pre-merge processing."""
        items = list(flat_config.dict.items())
        for flat_key, val in items:
            if is_tag_in(flat_key, "select"):
                # Remove the tag
                clean_key = clean_all_tags(flat_key)
                del flat_config.dict[flat_key]
                flat_config.dict[clean_tag(flat_key, "select")] = val
                self.keys_that_select.add(clean_key)
                if isinstance(val, str):
                    subconfig = ".".join(flat_key.split(".")[:-1])
                    keys_to_keep = [clean_key, val]
                elif isinstance(val, list):
                    subconfig = ".".join(val[0].split(".")[:-1])
                    for key in val[1:]:
                        subconfig2 = ".".join(key.split(".")[:-1])
                        if subconfig != subconfig2:
                            raise ValueError(
                                "The keys in the list of parameters tagged with "
                                "'@select' must be identical before the last dot "
                                f"(= on the same subconfig). Find: {subconfig} and "
                                f"{subconfig2} before the last dot."
                            )
                    keys_to_keep = [clean_key] + val
                else:
                    raise ValueError(
                        "The value of parameters tagged with '@select' must be a "
                        "string or a list of strings representing flat key(s). "
                    )
                subconfig = clean_all_tags(subconfig)
                if subconfig == "":
                    raise ValueError(
                        "Find attempt to delete the configuration at the root. You "
                        "must pass a flat key with a least one dot on parameter "
                        f"tagged with @select. Find key: {flat_key} with value: {val}"
                    )
                self.subconfigs_to_delete.add(subconfig)
                self.keys_to_keep.update(keys_to_keep)
        return flat_config

    def postmerge(self, flat_config: Config) -> Config:
        """Post-merge processing."""
        def _is_in_subconfig(key: str, subconfig: str) -> bool:
            """Check if a key is in a subconfig with the exact name."""
            return key == subconfig or key.startswith(subconfig + ".")
        # Delete all keys on the subconfigs except the ones to keep
        for subconfig in self.subconfigs_to_delete:
            for key in list(flat_config.dict.keys()):
                if (_is_in_subconfig(key, subconfig)
                        and not any(_is_in_subconfig(key, key_to_keep)
                                    for key_to_keep in self.keys_to_keep)):
                    del flat_config.dict[key]
        return super().postmerge(flat_config)

    def presave(self, flat_config: Config) -> Config:
        """Pre-save processing."""
        # Restore the tag with the type to keep the information
        # on further loading
        for key in self.keys_that_select:
            if key in flat_config.dict:
                new_key = key + "@select"
                flat_config.dict[new_key] = flat_config.dict[key]
                del flat_config.dict[key]
        return flat_config


class ProcessCheckTags(Processing):
    """Raise an error if a tag is present in a key after pre-merging processes.

    This security processing is applied after all pre-merge process and
    checks for "@" in the keys. It raises an error if one is found.
    """

    def __init__(self) -> None:
        super().__init__()
        # NOTE: this processing is a special meta-processing that must be
        # applied after all other pre-merge processing to ensure security.
        # That why it has a very high pre-merge order and it is not a
        # good idea to make pre-merge processing with higher order.
        self.premerge_order = 1000.0

    def premerge(self, flat_config: Config) -> Config:
        """Post-merge processing."""
        _, tagged_keys = dict_clean_tags(flat_config.dict)
        if tagged_keys:
            keys_message = "\n".join(tagged_keys[:5])
            raise ValueError(
                "Keys with tags are encountered at the end of "
                "the pre-merge process. It is probably a mistake due to:\n"
                "- a typo in tag name\n"
                "- a missing processing in process_list\n"
                "- the use of an 'at' ('@') in a parameter name\n"
                "- the use of a custom processing that does not remove a tag\n\n"
                "The tagged keys encountered (5 first if more than 5) are:\n"
                f"{keys_message}"
            )
        return flat_config


class DefaultProcessings():
    """Default list of built-in processings.

    To add these processings to a Config instance, use:

    config.process_list += DefaultProcessings().list

    The current default processing list contains:
     * ProcessCheckTags: protect against '@' in keys at the end of pre-merge)
     * ProcessMerge (@merge_all, @merge_before, @merge_after): merge multiple
       files into one.
     * ProcessCopy (@copy): persistently copy a value from one key to an other
       and protect it
     * ProcessTyping (@type:X): force the type of parameter to any type X.
     * ProcessSelect (@select): select sub-config(s) to keep and delete the
       other sub-configs in the same parent config.
    """

    def __init__(self) -> None:
        self.list: List[Processing] = [
            ProcessCheckTags(),
            ProcessMerge(),
            ProcessCopy(),
            ProcessTyping(),
            ProcessSelect()
        ]



"""Test for cli_parser.py."""
import pytest
import pytest_check as check

from cliconfig.cli_parser import parse_cli


def test_parse_cli() -> None:
    """Test for parse_cli."""
    config_paths, config_cli_params = parse_cli(
        ["main.py", "--config", "config1.yaml,config2.yaml", "--a=1", "--b=2"]
    )
    check.equal(config_paths, ["config1.yaml", "config2.yaml"])
    config_paths, config_cli_params = parse_cli(
        ["main.py", "--config", "[config1.yaml, config2.yaml]", "--a=1", "--b=2"]
    )
    check.equal(config_paths, ["config1.yaml", "config2.yaml"])
    check.equal(config_cli_params, {"a": 1, "b": 2})
    b_str = "{'2': {'3': 4, '5': [6.3], '7': True, '8': Null, '9': [2, {'a': 1}]}}"
    config_paths, config_cli_params = parse_cli(
        ["main.py", "--a='b=c'", f"--b={b_str}"]
    )
    val = {
        "a": "b=c",
        "b": {"2": {"3": 4, "5": [6.3], "7": True, "8": None, "9": [2, {"a": 1}]}},
    }
    check.equal(config_paths, [])
    check.equal(config_cli_params, val)
    config_paths, config_cli_params = parse_cli(
        ["main.py", "--unknown", "--a=1", "--unknwon2=2", "--b=3"]
    )
    val2 = {"unknown": None, "a": 1, "unknwon2": 2, "b": 3}
    check.equal(config_paths, [])
    check.equal(config_cli_params, val2)
    with pytest.raises(ValueError, match="Only one '--config ' argument is allowed.*"):
        parse_cli(["main.py", "--config", "config1.yaml", "--config", "config2.yaml"])



"""Base class for processing."""
from cliconfig.base import Config


class Processing():
    """Processing base class."""

    def __init__(self) -> None:
        self.premerge_order = 0.0
        self.postmerge_order = 0.0
        self.presave_order = 0.0
        self.postload_order = 0.0

    def premerge(self, flat_config: Config) -> Config:
        """Pre-merge processing.

        Function that can be applied to the flat config to modify it
        before merging . It takes a flat config and returns a flat config.
        """
        return flat_config

    def postmerge(self, flat_config: Config) -> Config:
        """Post-merge processing.

        Function that can be applied to the flat config to modify it
        after merging . It takes a flat config and returns a flat config.
        """
        return flat_config

    def presave(self, flat_config: Config) -> Config:
        """Pre-save processing.

        Function applied to the flat config to modify it before
        saving it. It takes a flat config and returns a flat config.
        """
        return flat_config

    def postload(self, flat_config: Config) -> Config:
        """Post-load processing.

        Function applied to the flat config to modify it after
        loading it. It takes a flat config and returns a flat config.
        """
        return flat_config

    def __eq__(self, __value: object) -> bool:
        """Equality operator.

        Two processing are equal if they are the same class and add the same
        attributes (get with __dict__).
        """
        equal = (
            isinstance(__value, self.__class__) and self.__dict__ == __value.__dict__
        )
        return equal



"""Private module with type parser for processing module with type manipulation."""
from pydoc import locate
from typing import List, Optional, Tuple, Type, Union


def _parse_type(type_desc: str) -> Tuple:
    """Parse a type description.

    Allow basic types (none, any, bool, int, float, str, list, dict), nested lists,
    nested dicts, unions (with Union or the '|' symbol) and Optional.

    Representation:
     * None|Any -> (type(None), object)
     * list|float -> (list, float)
     * List[float] -> (("list", (float,)),)
     * Dict[str, List[float]] -> (("dict", (str,), ((("list", (float,)),),)),)
     * Dict[str|bool, int|float] -> (("dict", (str, bool), (int, float)),)

    Raises
    ------
    ValueError
        If the type description is not valid or not recognized.
    """
    # Clean up
    old_type_desc = type_desc
    type_desc = type_desc.replace(' ', '').lower()
    try:
        # Base typeParseType
        base_type = _parse_base_type(type_desc)
        if base_type is not None:
            return (base_type, )
        # Split the external blocks enclosed by brackets
        blocks = _split_brackets(type_desc, delimiter='|')
        types: Tuple = ()
        for block in blocks:
            if block[:5] == 'list[':
                types += _parse_list(block)
            elif block[:5] == 'dict[':
                types += _parse_dict(block)
            elif block[:9] == 'optional[':
                types += _parse_optional(block)
            elif block[:6] == 'union[':
                types += _parse_union(block)
            else:  # Should be a base type
                base_type = _parse_base_type(block)
                if base_type is not None:
                    types += (base_type, )
                else:
                    raise ValueError(f"Unknown type: '{block}'")
        return types
    except ValueError as err:
        # Revert the traceback to show the original type description
        raise ValueError(f"Unknown type: '{old_type_desc}'") from err


def _parse_base_type(type_desc: str) -> Optional[Type]:
    """Parse a base type description.

    Base types are: none, any, bool, int, float and str, list, dict.
    Return None if the type is not a base type.
    """
    if type_desc == 'none':
        return type(None)
    if type_desc == 'any':
        # Match any type
        return object
    if type_desc in ('bool', 'int', 'float', 'str', 'list', 'dict'):
        return locate(type_desc)  # type: ignore
    return None


def _parse_list(type_desc: str) -> Tuple:
    """Parse a "list" type description."""
    sub_desc = type_desc[5:-1]
    if len(_split_brackets(sub_desc, delimiter=',')) > 1:
        raise ValueError(f"Invalid List type: '{type_desc}'")
    return (('list', ) + (_parse_type(sub_desc), ), )


def _parse_dict(type_desc: str) -> Tuple:
    """Parse an "dict" type description."""
    sub_desc = type_desc[5:-1]
    sub_blocks = _split_brackets(sub_desc, delimiter=',')
    if len(sub_blocks) != 2:
        raise ValueError(f"Invalid Dict type: '{type_desc}'")
    key_type = _parse_type(sub_blocks[0])
    value_type = _parse_type(sub_blocks[1])
    return (('dict', ) + (key_type, ) + ((value_type), ), )


def _parse_optional(type_desc: str) -> Tuple:
    """Parse an "optional" type description."""
    sub_desc = type_desc[9:-1]
    if len(_split_brackets(sub_desc, delimiter=',')) > 1:
        raise ValueError(f"Invalid Optional type: '{type_desc}'")
    return ((type(None), ) + _parse_type(sub_desc), )


def _parse_union(type_desc: str) -> Tuple:
    sub_desc = type_desc[6:-1]
    # Split by comma
    sub_blocks = _split_brackets(sub_desc, delimiter=',')
    if len(sub_blocks) < 2:
        raise ValueError(f"Invalid Union type: '{type_desc}'")
    types: Tuple = ()
    union_types = [_parse_type(sub_block) for sub_block in sub_blocks]
    for union_type in union_types:
        types += union_type
    return types


def _split_brackets(type_desc: str, delimiter: str) -> List[str]:
    """Split a type description in blocks enclosed by brackets."""
    blocks = []
    bracket_count = 0
    i, j = 0, 0
    while j < len(type_desc):
        char = type_desc[j]
        if char == '[':
            bracket_count += 1
        elif char == ']':
            bracket_count -= 1
        if bracket_count == 0 and char == delimiter:
            blocks.append(type_desc[i:j])
            i = j + 1
            j += 2
        else:
            j += 1
    blocks.append(type_desc[i:j])
    return blocks


def _isinstance(obj: object, types: Union[Type, Tuple]) -> bool:
    """Check if an object is an instance of a type or a tuple of types.

    Intended to work with the outputs of _parse_type.
    """
    if isinstance(types, type):
        return isinstance(obj, types)
    if types[0] == 'list' and len(types) == 2:
        return (isinstance(obj, list)
                and all(_isinstance(elem, types[1]) for elem in obj))
    if types[0] == 'dict' and len(types) == 3:
        return (isinstance(obj, dict)
                and all(_isinstance(key, types[1]) for key in obj)
                and all(_isinstance(value, types[2]) for value in obj.values()))
    if isinstance(types[0], (type, tuple)):
        return any(_isinstance(obj, sub_types) for sub_types in types)
    raise ValueError(f"Invalid type for _isinstance: '{types}'")



"""Test base class of processing."""

import pytest_check as check

from cliconfig.base import Config
from cliconfig.processing.base import Processing


def test_processing() -> None:
    """Test Processing."""
    class _ProcessingTest(Processing):

        def __init__(self) -> None:
            super().__init__()
            self.attr = 0

    config = Config({"a.b": 1, "b": 2, "c.d.e": 3, "c.d.f": [2, 3]}, [])
    base_process = Processing()
    check.equal(
        (
            base_process.premerge(config),
            base_process.postmerge(config),
            base_process.presave(config),
            base_process.postload(config),
        ),
        (config, config, config, config),
    )
    # Check equality of Processing objects
    proc1 = _ProcessingTest()
    proc1.attr = 0
    proc2 = _ProcessingTest()
    proc2.attr = 0
    check.equal(proc1, proc2)
    proc2.attr = 1
    check.not_equal(proc1, proc2)



"""Routines to manipulate dictionaries with processing."""
from typing import List, Optional, Union

from cliconfig.base import Config
from cliconfig.dict_routines import (
    _flat_before_merge,
    flatten,
    load_dict,
    merge_flat,
    save_dict,
    unflatten,
)
from cliconfig.processing.base import Processing


def merge_flat_processing(
    config1: Config,
    config2: Config,
    *,
    allow_new_keys: bool = True,
    preprocess_first: bool = True,
    preprocess_second: bool = True,
    postprocess: bool = True,
) -> Config:
    """Flatten, merge config2 into config1 and apply pre and post processing.

    Work even if the config dicts have a mix of nested and flat dictionaries.
    If both arguments are configs, the process lists are merged before applying
    the processing. The duplicate processings are removed.

    Parameters
    ----------
    config1 : Config
        The first config.
    config2 : Config
        The second dict to merge into config1.
    allow_new_keys : bool, optional
        If True, new keys (that are not in config1) are allowed in config2.
        Otherwise, it raises an error. By default True.
    preprocess_first : bool, optional
        If True, apply pre-merge processing to config1. By default True.
    preprocess_second : bool, optional
        If True, apply pre-merge processing to config2. By default True.
    postprocess : bool, optional
        If True, apply post-merge processing to the merged config. By default True.

    Raises
    ------
    ValueError
        If allow_new_keys is False and config2 has new keys that are not in config1.
    ValueError
        If there are conflicting keys when flatten one of the dicts.

    Returns
    -------
    flat_config : Config
        The merged flat config.
    """
    # Flatten the dictionaries
    config1.dict, config2.dict = _flat_before_merge(config1.dict, config2.dict)
    # Get the process list of the merge
    process_list = config1.process_list
    for process in config2.process_list:
        # NOTE 2 processings are equal if they are the same class and add the same
        # attributes.
        if process not in process_list:
            process_list.append(process)
    # Apply the pre-merge processing
    if preprocess_first:
        config1.process_list = process_list
        pre_order_list = sorted(process_list, key=lambda x: x.premerge_order)
        for processing in pre_order_list:
            config1 = processing.premerge(config1)
        process_list = config1.process_list
    if preprocess_second:
        config2.process_list = process_list
        pre_order_list = sorted(process_list, key=lambda x: x.premerge_order)
        for processing in pre_order_list:
            config2 = processing.premerge(config2)
        process_list = config2.process_list
    # Merge the dictionaries
    flat_dict = merge_flat(config1.dict, config2.dict, allow_new_keys=allow_new_keys)
    # Create the new config
    flat_config = Config(flat_dict, process_list)
    # Apply the postmerge processing
    if postprocess:
        post_order_list = sorted(process_list, key=lambda x: x.postmerge_order)
        for processing in post_order_list:
            flat_config = processing.postmerge(flat_config)
    return flat_config


def merge_flat_paths_processing(
    config_or_path1: Union[str, Config],
    config_or_path2: Union[str, Config],
    *,
    additional_process: Optional[List[Processing]] = None,
    allow_new_keys: bool = True,
    preprocess_first: bool = True,
    preprocess_second: bool = True,
    postprocess: bool = True,
) -> Config:
    """Flatten, merge and apply processing to two configs or their yaml paths.

    Similar to :func:`merge_flat_processing` but allows to pass configs
    or yaml paths. Work even if the configs have a mix of nested and flat dicts.
    If both arguments are configs, the process lists are merged before applying
    the processing. The duplicate processings are removed.

    Parameters
    ----------
    config_or_path1 : Union[str, Config]
        The first config or its path.
    config_or_path2 : Union[str, Config]
        The second config or its path, to merge into first config.
    additional_process : Optional[List[Processing]], optional
        Additional processings to apply to the merged config. It can
        be useful to merge a config from its path while it has some specific
        processings.
    allow_new_keys : bool, optional
        If True, new keys (that are not in config1) are allowed in config2.
        Otherwise, it raises an error. By default True.
    preprocess_first : bool, optional
        If True, apply pre-merge processing to config1. By default True.
    preprocess_second : bool, optional
        If True, apply pre-merge processing to config2. By default True.
    postprocess : bool, optional
        If True, apply post-merge processing to the merged config. By default True.

    Raises
    ------
    ValueError
        If allow_new_keys is False and config2 has new keys that are not in config1.
    ValueError
        If there are conflicting keys when flatten one of the dicts.

    Returns
    -------
    flat_config : Config
        The merged flat config.
    """
    configs = []
    for config_or_path in [config_or_path1, config_or_path2]:
        if isinstance(config_or_path, str):
            config_dict = load_dict(config_or_path)
            config = Config(config_dict, [])
        elif isinstance(config_or_path, Config):
            config = config_or_path
        elif isinstance(config_or_path, dict):
            raise ValueError(
                "config_or_path must be a Config instance or a path to a yaml file "
                "but you passed a dict. If you want to use it as a valid input, "
                "you should use Config(<input dict>, []) instead."
            )
        else:
            raise ValueError(
                "config_or_path must be a Config instance or a path to a yaml file."
            )
        configs.append(config)
    config1, config2 = configs[0], configs[1]
    if additional_process is not None:
        config1.process_list.extend(additional_process)
        config2.process_list.extend(additional_process)
    flat_config = merge_flat_processing(
        config1,
        config2,
        allow_new_keys=allow_new_keys,
        preprocess_first=preprocess_first,
        preprocess_second=preprocess_second,
        postprocess=postprocess,
    )
    return flat_config


def save_processing(config: Config, path: str) -> None:
    """Save a config and apply pre-save processing before saving.

    Parameters
    ----------
    config : Config
        The config to save.
    path : str
        The path to the yaml file to save the config dict.
    """
    config.dict = flatten(config.dict)
    # Get the pre-save order
    order_list = sorted(config.process_list, key=lambda x: x.presave_order)
    # Apply the pre-save processing
    for processing in order_list:
        config = processing.presave(config)
    # Unflatten and save the dict
    config.dict = unflatten(config.dict)
    save_dict(config.dict, path)


def load_processing(path: str, process_list: List[Processing]) -> Config:
    """Load a dict from yaml file path and apply post-load processing.

    Parameters
    ----------
    path : str
        The path to the file to load the dict.
    process_list: List[Processing]
        The list of processing to apply after loading. Only post-load
        processing is applied. The order of the processing is given
        by the postload_order attribute of the processing.

    Returns
    -------
    flat_config : Config
        The loaded flat config.
    """
    # Load the dict and flatten it
    out_dict = flatten(load_dict(path))
    flat_config = Config(out_dict, process_list)
    # Get the post-load order
    order_list = sorted(process_list, key=lambda x: x.postload_order)
    # Apply the post-load processing
    for processing in order_list:
        flat_config = processing.postload(flat_config)
    return flat_config



"""Parser for CLI commands."""
from typing import Any, Dict, List, Tuple

import yaml


def parse_cli(sys_argv: List[str]) -> Tuple[List[str], Dict[str, Any]]:
    """Parser for CLI commands.

    Return list of config paths that are detected with `--config ` (with a space),
    separated by commas without space (or in a list).
    Return also a dictionary of parameters from CLI detected with --<key>=<value>
    (with "=" and without spaces).

    Parameters
    ----------
    sys_argv : List[str]
        List of arguments from sys.argv.

    Raises
    ------
    Value Error
        If the '--config ' argument (with space) is used more than once.

    Returns
    -------
    config_paths : List[str]
        List of paths to config files to merge.
    cli_params_dict : Dict[str, Any]
        Dictionary of parameters from CLI.

    Examples
    --------
    .. code-block:: text

        $ python my_script.py --config config.yaml --foo.bar.param=[1, 2, 3]

    Will be parsed as `config_paths=['config.yaml']`
    and `cli_params={'foo.bar.param': [1, 2, 3]}`.
    It is equivalent to: `{'foo': {'bar': {'param': [1, 2, 3]}}`
    for :func:`merge_config`.
    """
    cli_params_dict: Dict[str, Any] = {}
    config_paths: List[str] = []
    i = 0
    while i < len(sys_argv):
        elem = sys_argv[i]
        if elem == "--config":
            if config_paths:
                raise ValueError(
                    "Only one '--config ' argument is allowed in CLI (used for "
                    "config merging)."
                )
            configs = yaml.safe_load(sys_argv[i + 1])
            if isinstance(configs, list):
                config_paths = configs
            if isinstance(configs, str):
                config_paths = sys_argv[i + 1].split(",")
            i += 2
        elif elem.startswith("--"):
            splits = elem.split("=", maxsplit=1)
            if len(splits) == 2:
                key, value_str = splits
            else:
                key = splits[0]
                value_str = "null"
            key = key[2:]
            value = yaml.safe_load(value_str)
            cli_params_dict[key] = value
            i += 1
        else:  # Not a config parameter
            i += 1
    return config_paths, cli_params_dict



"""Base classes."""
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, List

# Imports Processing for mypy only
if TYPE_CHECKING:
    from cliconfig.processing.base import Processing


@dataclass
class Config():
    """Class for configuration. It stores the config dict and the processing list."""

    dict: Dict[str, Any]  # noqa: A003
    process_list: List["Processing"]



"""Manage Pytest-cov output on workflow."""
import sys

from github_actions_utils.color import score_to_hex_color


def check_output() -> float:
    """Check output of Pytest-cov.

    Raises
    ------
    ValueError
        If Pytest find failures.
    ValueError
        If coverage is below SCORE_MIN.

    Returns
    -------
    score: float
        Score of coverage.
    """
    args = sys.argv[1:]
    n_failures, score = -1, -1.0  # Default values
    for arg in args:
        if arg.startswith("--score="):
            score_percent = arg.split("=")[1]
            score = float(score_percent.split("%")[0])
        if arg.startswith("--n_failures="):
            n_failures_str = arg.split("=")[1]
            n_failures = 0 if n_failures_str == "" else int(n_failures_str)

    if n_failures == -1:
        raise ValueError(
            "Please specify the number of failures found by Pytest "
            "using the flag --n_failures=N.",
        )
    if score == -1:
        raise ValueError(
            "Please specify the score of coverage using the flag --score=N (in %).",
        )

    if n_failures > 0:
        raise ValueError(f"Pytest finds {n_failures} failure(s) on tests.")
    if score < COV_SCORE_MIN:
        raise ValueError(
            f"Pytest coverage {score}% is lower than minimum ({COV_SCORE_MIN}%)",
        )

    return score


def main() -> None:
    """Check score and print it."""
    score = check_output()
    # Print color to be used in GitHub Actions
    print(score_to_hex_color(score, COV_SCORE_MIN, COV_SCORE_MAX))


if __name__ == "__main__":
    # COV_SCORE_MIN can be changed safely depending on your needs.
    # NOTE: score on %
    COV_SCORE_MIN = 0
    COV_SCORE_MAX = 100
    main()



"""Color utils for GitHub actions."""
from colorsys import hsv_to_rgb


def score_to_hex_color(score: float, score_min: float, score_max: float) -> str:
    """Convert score to hex color red > bright green."""
    norm_score = max(0, (score - score_min) / (score_max - score_min))
    hsv = (1 / 3 * norm_score, 1, 1)
    rgb = hsv_to_rgb(*hsv)
    rgb_tuple = tuple(int(255 * value) for value in rgb)
    hex_color = f"#{rgb_tuple[0]:02x}{rgb_tuple[1]:02x}{rgb_tuple[2]:02x}"
    return hex_color



"""Functions to manipulate config as dict with yaml files and CLI."""
import sys
from typing import List, Optional

from cliconfig.base import Config
from cliconfig.cli_parser import parse_cli
from cliconfig.dict_routines import flatten, show_dict, unflatten
from cliconfig.process_routines import (
    load_processing,
    merge_flat_paths_processing,
    merge_flat_processing,
    save_processing,
)
from cliconfig.processing.base import Processing
from cliconfig.processing.builtin import DefaultProcessings
from cliconfig.tag_routines import clean_all_tags


def make_config(
    *default_config_paths: str,
    process_list: Optional[List[Processing]] = None,
    add_default_processing: bool = True,
) -> Config:
    r"""Make a config from default configs and CLI arguments and apply processing.

    It uses the CLI Config routines to parse the CLI arguments and merge
    them in a simple way. Apply the processing functions on each merge.

    Parameters
    ----------
    default_config_paths : Tuple[str]
        Paths to default configs. They are merged in order and new keys
        are allowed.
    process_list: Optional[List[Processing]], optional
        The list of processing to apply during each merge. None for empty list.
        By default None.
        If add_default_processing is True, the default processing
        (in :mod:`cliconfig.processing.builtin`) are added to the list.
    add_default_processing : bool, optional
        If True, add the default processing (in :mod:`cliconfig.processing.builtin`)
        to the processing list.

    Raises
    ------
    ValueError
        If allow_new_keys is False and CLI has new keys that are not
        in default configs.

    Returns
    -------
    config : Config
        The built config. Contains the config dict (config.dict) and the processing
        list (config.process_list), which can be used to apply further processing
        routines.

    Examples
    --------
    ::

        # main.py
        config = make_config('data.yaml', 'model.yaml', 'train.yaml')

    .. code-block:: text

        python main.py -- config bestmodel.yaml,mydata.yaml \
            --architecture.layers.hidden_dim=64

    """
    # Create the processing list
    process_list_: List[Processing] = [] if process_list is None else process_list
    if add_default_processing:
        process_list_ += DefaultProcessings().list

    config = Config({}, process_list_)

    # Merge default configs and additional configs
    additional_config_paths, cli_params_dict = parse_cli(sys.argv)
    for i, paths in enumerate([default_config_paths, additional_config_paths]):
        # Allow new keys for default configs only
        allow_new_keys = i == 0
        for path in paths:
            config = merge_flat_paths_processing(
                config,
                path,
                allow_new_keys=allow_new_keys,
                preprocess_first=False,  # Already processed
            )

    # Allow new keys for CLI parameters but do not merge them and raise
    # warning.
    cli_params_dict = flatten(cli_params_dict)
    new_keys, keys = [], list(cli_params_dict.keys())
    for key in keys:
        if clean_all_tags(key) not in config.dict:
            # New key: delete it
            new_keys.append(clean_all_tags(key))
            del cli_params_dict[key]
    if new_keys:
        new_keys_message = "  - " + "\n  - ".join(new_keys)
        print(
            "[CONFIG] Warning: New keys found in CLI parameters "
            f"that will not be merged:\n{new_keys_message}"
        )
    # Merge CLI parameters
    cli_params_config = Config(cli_params_dict, [])
    config = merge_flat_processing(
        config,
        cli_params_config,
        allow_new_keys=False,
        preprocess_first=False
    )
    print(
        f"[CONFIG] Info: Merged {len(default_config_paths)} default config(s), "
        f"{len(additional_config_paths)} additional config(s) and "
        f"{len(cli_params_dict)} CLI parameter(s)."
    )
    # Unflatten the config dict
    config.dict = unflatten(config.dict)
    return config


def load_config(
    path: str,
    default_config_paths: Optional[List[str]] = None,
    process_list: Optional[List[Processing]] = None,
    *,
    add_default_processing: bool = True,
) -> Config:
    """Load config from a file and merge into optional default configs.

    First merge the default configs together, then load the default config,
    apply the post-load processing, and finally merge the loaded config

    Parameters
    ----------
    path : str
        The path to the file to load the configuration.
    default_config_paths : Optional[List[str]], optional
        Paths to default configs. They are merged in order, new keys are allowed.
        Then, the loaded config is merged into the result. None for no default configs.
        By default None.
    process_list: Optional[List[Processing]]
        The list of processing to apply after loading. Only postload
        method is applied. The order of the processing is given
        by the postload_order attribute of the processing.
        If None, no processing is applied. By default None.
    add_default_processing : bool, optional
        If True, add the default processing to the processing list.

    Returns
    -------
    config: Dict[str, Any]
        The loaded config. Contains the config dict (config.dict) and the processing
        list (config.process_list), which can be used to apply further processing
        routines.
    """
    # Crate process_list
    process_list_: List[Processing] = [] if process_list is None else process_list
    if add_default_processing:
        process_list_ += DefaultProcessings().list

    config = Config({}, process_list_)
    if default_config_paths:
        for config_path in default_config_paths:
            config = merge_flat_paths_processing(
                config,
                config_path,
                allow_new_keys=True,
                preprocess_first=False,  # Already processed
            )
    # Disallow new keys for loaded config
    loaded_config = load_processing(path, config.process_list)
    # Update the config list from loaded_config in config
    config.process_list = loaded_config.process_list
    # Merge the loaded config into the config
    # NOTE: The loaded config is processed with pre-merge so that tags in
    # the yaml files are correctly processed.
    config = merge_flat_processing(
        config,
        loaded_config,
        allow_new_keys=False,
        preprocess_first=False,
    )
    # Unflatten the config
    config.dict = unflatten(config.dict)
    return config


def save_config(config: Config, path: str) -> None:
    """Save a config and apply pre-save processing before saving.

    Parameters
    ----------
    config : Dict[str, Any]
        The config to save.
    path : str
        The path to the yaml file to save the dict.
    """
    save_processing(config, path)


def show_config(config: Config) -> None:
    """Show the config dict in a pretty way.

    The config dict is unflattened before.

    Parameters
    ----------
    config : Config
        The config to show.
    """
    print("Config:")
    show_dict(config.dict, start_indent=1)
